#include <cstdlib>
#include <iostream>
#include <vector>
#include <sstream>
#include <climits>
#include <time.h>
#include <math.h>
#include <pthread.h>

#include "mersenne.hpp"
#include "bmp.hpp"

// Width and height of the render
#define WIDTH       1050
#define HEIGHT      1050
// Center coordinates and zoom factor.
// I use factors and not quotients for the zoom, because it's easier to
// navigate through the fractal by multiples of the zoom factor.
#define CENTER_X    -0.4
#define CENTER_Y    0.0
#define ZOOM        0.7
// Mode selection
enum F_MODES
{
    MANDELBROT,
    BUDDHABROT
};
#define MODE        MANDELBROT
// Activate print output
#define VERBOSE     true
// Maximum iteration count (escape time)
#define MAX_TIME    10000
// Buddhabrot: minimum iteration count
#define MIN_TIME    0
// Buddhabrot: number of samples, in millions
#define N_SAMPLES   300
// Buddhabrot: coloration escape times
#define R_MIN       300
#define R_MAX       1000
#define G_MIN       75
#define G_MAX       300
#define B_MIN       0
#define B_MAX       75
// Mandelbrot: coloration mode
enum C_MODES
{
    GRAYSCALE,
    RAINBOW
};
#define COLORATION  RAINBOW
// Coloration factors, for both mandelbrot and buddhabrot
#define R_FACTOR    1.0
#define G_FACTOR    1.0
#define B_FACTOR    1.0
// Rainbow frequency
#define FREQUENCY   1.03

/*
 * Convention used:
 *   x , y  : real and imaginary parts
 *   x0, y0 : scaled coordinates in complex plane
 */

// Optimized escape-time formula
int escapeTime(double x0, double y0)
{
    // Check if point is in main cardioid or period-2 bulb
    double q = (x0 - 0.25)*(x0 - 0.25) + y0*y0;
    q = q*(q + (x0 - 0.25));
    if ((q < 0.25*y0*y0) || (((x0 + 1)*(x0 + 1) + y0*y0) < 0.0625))
        return MAX_TIME;
    // If not, iterate
    double x = 0;
    double y = 0;
    int iter = 0;
    while ((x*x + y*y <= 4) && (iter < MAX_TIME))
    {
        double new_x = x*x - y*y + x0;
        y = 2*x*y + y0;
        x = new_x;
        iter++;
    }
    return iter;
}
// Struct of the arguments to pass to the threads
struct mandel_args {
    int    y_start;
    int      y_end;
    double scale_x;
    double scale_y;
    int *count_ptr;
    std::vector<std::vector<px_24> > *pict_ptr;
};
// Method called by the threads
void* mandelbrot(void* args_ptr)
{
    // Parse args
    mandel_args *args = (mandel_args*) args_ptr;
    int y_start = args->y_start;
    int y_end   = args->y_end;
    double scale_x = args->scale_x;
    double scale_y = args->scale_y;
    int *count_ptr = args->count_ptr;
    std::vector<std::vector<px_24> > *pict_ptr = args->pict_ptr;

    // We proceed by lines
    for (int i = y_start; i < y_end; i++)
    {
        std::vector<px_24> line;
        
        for (int j = 0; j < WIDTH; j++)
        {
            // We must scale the discrete pixel coordinates
            // We apply the zoom and position here
            double x0 = ((1.*j / WIDTH)  * scale_x * ZOOM)
                        - (scale_x * ZOOM / 2) + CENTER_X;
            double y0 = ((1.*i / HEIGHT) * scale_y * ZOOM)
                        - (scale_y * ZOOM / 2) + CENTER_Y;
            int iter = escapeTime(x0, y0);
            // Coloration stuff
            px_24 pixel;
            if (COLORATION == GRAYSCALE)
            {
                // Intensity
                double c = (1.*iter / MAX_TIME) - 1;
                // Small color correction with -(x-1)^4 + 1 curve
                // for enhancing details
                uint8_t color = (uint8_t) ((-1*c*c*c*c + 1) * 0.97 * 255);
                // Apply color factors
                pixel.R = (uint8_t) (color * R_FACTOR);
                pixel.G = (uint8_t) (color * G_FACTOR);
                pixel.B = (uint8_t) (color * B_FACTOR);
            }
            else if (COLORATION == RAINBOW)
            {
                pixel.R = (uint8_t) ((sin(FREQUENCY*iter + 0) * 127 + 128) * R_FACTOR);
                pixel.G = (uint8_t) ((sin(FREQUENCY*iter + 2) * 127 + 128) * G_FACTOR);
                pixel.B = (uint8_t) ((sin(FREQUENCY*iter + 4) * 127 + 128) * B_FACTOR);
            }
            line.push_back(pixel);
            // Update the progress
            
            if (VERBOSE)
            {
                *count_ptr+=1;
            }
        }
        pict_ptr->push_back(line);
    }
}

int main(int argc, char** argv)
{
    // Detect CPUs
    int numCPU;
#ifdef _WIN32
    SYSTEM_INFO sysinfo;
    GetSystemInfo(&sysinfo);
    numCPU = sysinfo.dwNumberOfProcessors;
#else
    numCPU = sysconf( _SC_NPROCESSORS_ONLN );
#endif
    // Scale factors to normalize the discrete pixel coordinates
    // While 3.5 is arbitrary, SCALE_Y is calculated to keep
    // a correct aspect ratio.
    const double SCALE_X = 3.5;
    const double SCALE_Y = HEIGHT * SCALE_X / WIDTH;
    // Storage vectors for both pictures
    std::vector<std::vector<px_24> >    mandel_pict;
    std::vector<std::vector<px_float> > buddha_data;
    std::vector<std::vector<px_24> >    buddha_pict;
    // Print infos
    if (VERBOSE)
    {
        std::string mode = "";
        if (MODE == MANDELBROT) mode += "Mandelbrot";
        else mode += "Buddhabrot";
        std::cout
        << "Bread - Current mode: " << mode                         << std::endl
        << "----------------------------------------------"         << std::endl
        << "Width:\t" << WIDTH    << "\t\tHeight:\t\t" << HEIGHT    << std::endl
        << "Zoom:\t"  << ZOOM     << "\t\tIter.:\t\t"  << MAX_TIME  << std::endl
        << "X:\t"     << CENTER_X << "\t\tY:\t\t"        << CENTER_Y  << std::endl
        << std::endl;
    }
    // Keep track of time
    const int BEGIN_TIME = time(NULL);

    if (MODE == MANDELBROT)
    {
        std::vector<pthread_t> threads;
        std::vector<std::vector<std::vector<px_24> >* > pictures;
        int count = 0;
        int *count_ptr = &count;

        // Multi-threading
        numCPU = 2;
        for (int i = 0; i < numCPU; i++)
        {
            std::vector<std::vector<px_24> > *pict_ptr = new std::vector<std::vector<px_24> >();
            mandel_args args;
            args.y_start   = i     * (HEIGHT / numCPU);
            args.y_end     = (i+1) * (HEIGHT / numCPU);
            if (i == numCPU - 1)
                args.y_end = HEIGHT;
            args.scale_x   = SCALE_X;
            args.scale_y   = SCALE_Y;
            args.count_ptr = count_ptr;
            args.pict_ptr  = pict_ptr;
            pthread_t t1;
            pthread_create(&t1, NULL, &mandelbrot, (void*) &args);
            
            pictures.push_back(pict_ptr);
            std::cout << args.y_start << "->" << args.y_end << "\n";
            threads.push_back(t1);
        }
        // Output

        int step = (WIDTH*HEIGHT) / 100;
        // Following verbose block not working, because apparently when
        // Both threads increment count (by using a pointer) at the same time,
        // It skips..
        if (0) // if (VERBOSE)
        {
            while (count < WIDTH * HEIGHT)
            {
                if (count % step == 0) {
                    double p = (1. * count / (WIDTH * HEIGHT));
                    std::string bar((int) ((p * 44) - 1), '=');
                    std::string spc(43 - bar.size(), ' ');
                    std::cout << "[" << bar << ">" << spc << "] (" << (int) (p * 100) << "%)\r";
                }
            }
        }
        for (int i = 0; i < threads.size(); i++)
        {
            pthread_join(threads[i], NULL);
            for (int j = 0; j < pictures.at(i)->size(); j++)
            {
                mandel_pict.push_back(pictures.at(i)->at(j));
            }
        }
    }
    else if (MODE == BUDDHABROT)
    {
        int count = 0;
        // Init random seeds
        mt_init();
        // Fill data array with 0
        for (int i = 0; i < WIDTH; i++)
        {
            std::vector<px_float> column;
            for (int j = 0; j < HEIGHT; j++)
            {
                px_float p;
                p.R = 0;
                p.G = 0;
                p.B = 0;
                column.push_back(p);
            }
            buddha_data.push_back(column);
        }
        // We proceed by steps of millions of samples, by columns
        for (int i = 0; i < 1000000; i++)
        {
            for (int j = 0; j < N_SAMPLES; j++)
            {
                // We need two double between -2 and 2
                double x0  = 4*(1.*mt_random() / ULONG_MAX) - 2;
                double y0  = 4*(1.*mt_random() / ULONG_MAX) - 2;
                double x   = 0;
                double y   = 0;
                int iter   = 0;
                // If the random point is not in the mandelbrot set...
                bool inSet = (escapeTime(x0, y0) == MAX_TIME);
                // ... Draw the path it took to get out of it
                while ((!inSet) && (x*x + y*y <= 4) && (iter < MAX_TIME))
                {
                    double new_x = x*x - y*y + x0;
                    y = 2*x*y + y0;
                    x = new_x;
                    // Re-discretize the complex coordinates into pixel position
                    int d_x = (int) (WIDTH  * ((x + (SCALE_X * ZOOM / 2) - CENTER_X)) / (SCALE_X * ZOOM));
                    int d_y = (int) (HEIGHT * ((y + (SCALE_Y * ZOOM / 2) - CENTER_Y)) / (SCALE_Y * ZOOM));
                    // Check if it is displayable, and if it is past MIN_TIME
                    if (d_x >= 0 && d_y >= 0 && d_x < WIDTH && d_y < HEIGHT && iter > MIN_TIME)
                    {
                        // Do the right coloring according to settings
                        if (iter > R_MIN && iter < R_MAX)
                            buddha_data[d_x][d_y].R += 1;
                        if (iter > G_MIN && iter < G_MAX)
                            buddha_data[d_x][d_y].G += 1;
                        if (iter > B_MIN && iter < B_MAX)
                            buddha_data[d_x][d_y].B += 1;
                    }
                    iter++;
                }
            }
            if (VERBOSE)
            {
                count++;
                // Update progress bar very percent
                if (count % 10000 == 0)
                {
                    double p = (1.*count / 1000000);
                    std::string bar((int) ((p * 44) - 1), '=');
                    std::string spc(43 - bar.size(), ' ');
                    std::cout << "[" << bar << ">" << spc << "] (" << (int) (p * 100) << "%)\r";
                }
            }
        }
        // Now normalize and discretize every point
        // To normalize, we must now the maximum value of each color present
        float max_r = 0;
        float max_g = 0;
        float max_b = 0;
        for (int i = 0; i < buddha_data.size(); i++)
        {
            for (int j = 0; j < buddha_data[0].size(); j++)
            {
                px_float p = buddha_data[i][j];
                if (p.R > max_r) max_r = p.R;
                if (p.G > max_g) max_g = p.G;
                if (p.B > max_b) max_b = p.B;
            }
        }
        for (int i = 0; i < buddha_data.size(); i++)
        {
            // We proceed by columns
            std::vector<px_24> column;
            for (int j = 0; j < buddha_data[0].size(); j++)
            {
                // Compute every color (normalization), and take in account factors
                px_24 pixel;
                pixel.R = (uint8_t) ((buddha_data[i][j].R / max_r) * 255 * R_FACTOR);
                pixel.G = (uint8_t) ((buddha_data[i][j].G / max_g) * 255 * G_FACTOR);
                pixel.B = (uint8_t) ((buddha_data[i][j].B / max_b) * 255 * B_FACTOR);
                column.push_back(pixel);
            }
            buddha_pict.push_back(column);
        }
    }
    // Finish printing output
    if (VERBOSE)
    {
        std::string full(43, '=');
        std::cout << "[" << full << ">] (100%)" << std::endl
        << "Done in " << time(NULL) - BEGIN_TIME << " seconds, writing file... ";
    }
    // Prepare filename
    std::stringstream out;
    // Number after bread is the number of seconds since the first build :)
    out << "bread_" << time(NULL) - 1287759600
        << "_(X" << CENTER_X << ",Y" << CENTER_Y << ";" << ZOOM << ")";
    std::string mandel_name = out.str() + ".bmp";
    std::string buddha_name = out.str() + "_buddha.bmp";
    // Write files
    if (MODE == MANDELBROT) writeBmp(mandel_pict, mandel_name.c_str(), WIDTH, HEIGHT);
    if (MODE == BUDDHABROT) writeBmp(buddha_pict, buddha_name.c_str(), HEIGHT, WIDTH);
    if (VERBOSE)
    {
        std::cout << "Success !" << std::endl;
    }
    return EXIT_SUCCESS;
}
